/*****************************************************************************
 * container-ffmpeg.c : support for ffmpeg format output
 *****************************************************************************
 * $Id: container-ffmpeg.c,v 1.19 2007/02/19 15:04:27 pingus77 Exp $
 *****************************************************************************
 * Program under GNU General Public License (see ../COPYING)
 *****************************************************************************/

#include <config.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "container.h"
#include "divx.h"

#ifdef HAVE_FFMPEG
#if LIBAVCODEC_BUILD >= 4754
# include "avcodec.h"
# include "avformat.h"

static AVOutputFormat *av_output_fmt = NULL;
static AVFormatContext *oc=NULL; 
static AVStream *video_st=NULL;
static AVStream *audio_st=NULL;
static AVPacket pkt;

extern pthread_mutex_t avi_mutex;
extern AVCodecContext *codec_context;
extern AVCodecContext *audio_codec_context;

static  int open_ofile (struct CONTAINER* This, char* filename ,int maxsize)
{
    av_register_all();
    av_output_fmt = guess_format(This->shortname,NULL, NULL);

    if (!av_output_fmt) {
      fprintf(stderr, "Could not find suitable output format\n");
      return -1;
    }

    /* allocate the output media context */
    oc = av_alloc_format_context();
    if (!oc) {
	fprintf(stderr, "Memory error\n");
	return -1;
    }
    oc->oformat = av_output_fmt;
    snprintf(oc->filename, sizeof(oc->filename), "%s", filename);
    /* add the audio and video streams using the default format codecs
	       and initialize the codecs */
    video_st = NULL;
    audio_st = NULL;
    if (av_output_fmt->video_codec != CODEC_ID_NONE) {
//	int o_codec=-1;
/*	if (-1 == (o_codec =xdtv_to_ffmpeg(divx.codec) )){
		fprintf(stderr,"the video codec must be a ffmpeg codec ! \n");
		return NULL;
	}
*/	// we suppose that its ok...
	video_st = av_new_stream(oc,0); 
	video_st->codec = codec_context;	
    }


    if (av_output_fmt->audio_codec != CODEC_ID_NONE) {
//	int o_codec=-1;
/*	if (-1 == (o_codec = xdtv_to_audioffmpeg(divx.audio_codec) )){
		fprintf(stderr,"the audio codec must be a ffmpeg codec ! \n");
		return NULL;
	}
*/
	audio_st = av_new_stream(oc,0); 
	audio_st->codec = audio_codec_context;	
    }

    /* set the output parameters (must be done even if no
	       parameters). */
    if (av_set_parameters(oc, NULL) < 0) {
	fprintf(stderr, "Invalid output format parameters\n");
	return -1;
    }

    dump_format(oc, 0, filename, 1);

    /* open the output file, if needed */
    if (!(av_output_fmt->flags & AVFMT_NOFILE)) {
       	if (url_fopen(&oc->pb, filename, URL_WRONLY) < 0) {
       		fprintf(stderr, "Could not open '%s'\n", filename);
       		return -1;
       	}
    }
    
    /* write the stream header, if any */
    av_write_header(oc);

    /* now we just have to contruct an avi_t struct */
    if (NULL == (This->avfile = (avi_t *) malloc(sizeof(avi_t)))){
	    return -1;
    }
    memset((void *)This->avfile,0,sizeof(avi_t));
    This->avfile->init = 0;
    This->avfile->mode = AVI_MODE_WRITE; /* open for writing */
    This->avfile->seekable = 1;
    This->avfile->send_data = NULL;
    pthread_mutex_init (&avi_mutex, NULL);
    return 0;
}
static  int    write_frame(struct CONTAINER* This, char* data, long bytes, long flags)
{
  long pos;
  int ret=0;
  AVPacket pkt;
  pthread_mutex_lock(&avi_mutex);
  pos = This->avfile->pos;
  av_init_packet(&pkt);
  pkt.pts= av_rescale_q(video_st->codec->coded_frame->pts,video_st->codec->time_base, video_st->time_base);
  if(video_st->codec->coded_frame->key_frame)
	pkt.flags |= PKT_FLAG_KEY;
  pkt.stream_index= video_st->index;
  pkt.data= (uint8_t *) data;
  pkt.size= bytes;
   
  /* write the compressed frame in the media file */
  ret = av_write_frame(oc, &pkt);
  This->avfile->pos += bytes;
  /* and in the socket if any ... */
  if(This->avfile->send_data)
	This->avfile->send_data(data,bytes);

  if (ret != 0) {
    fprintf(stderr, "Error while writing video frame\n");
    return -1;
  }

  This->avfile->last_pos = pos;
  This->avfile->last_len = bytes;
  This->avfile->video_frames++;
  pthread_mutex_unlock(&avi_mutex);
  return 0;
}
static  int    write_audio(struct CONTAINER* This, char *data, long bytes)
{
   pthread_mutex_lock(&avi_mutex);
   av_init_packet(&pkt);
   pkt.size=bytes;
   pkt.pts= av_rescale_q(audio_st->codec->coded_frame->pts, audio_st->codec->time_base, audio_st->time_base);
   pkt.flags |= PKT_FLAG_KEY;
   pkt.stream_index= audio_st->index;
   pkt.data= (uint8_t*) data;

   /* write the compressed frame in the media file */
   if (av_write_frame(oc, &pkt) != 0) {
	fprintf(stderr, "Error while writing audio frame\n");
	exit(1);
   }
   This->avfile->pos += bytes;
   /* and in the socket if any ... */
   if(This->avfile->send_data)
	This->avfile->send_data(data,bytes);
   
   This->avfile->audio_bytes += bytes;
   pthread_mutex_unlock(&avi_mutex);
   return 0;
}

static  int 	 close_ofile      (struct CONTAINER* This)
{
  int i;
  pthread_mutex_destroy(&avi_mutex);
  /* write the trailer, if any */
  av_write_trailer(oc);
  /* free the streams */
  for(i = 0; i < oc->nb_streams; i++) {
    av_freep(&oc->streams[i]);
  }
  if (!(av_output_fmt->flags & AVFMT_NOFILE)) {
   /* close the output file */
    url_fclose(&oc->pb);
  }
  /* free the stream */
  av_free(oc);

  close(This->avfile->fdes);
  if(This->avfile->idx) free(This->avfile->idx);
  if(This->avfile->video_index) free(This->avfile->video_index);
  if(This->avfile->audio_index) free(This->avfile->audio_index);
  free(This->avfile);
  return 0;
}

static  unsigned 
  char*  get_header (struct CONTAINER* This, int* ret_size)
{
/* first, we write to a file like /tmp/xdtv_header_stream*/
    unsigned char* data=NULL;
    AVFormatContext header_oc;
    const char *file_header="/tmp/xdtv_header_stream";
    struct stat buf;
    int header_size;
    int headerfd;
    int size_read;

    /* copy the current output context */
    header_oc=*oc;
   
    /* open a new header file */
    fprintf(stderr,"create new header stream : %s\n",file_header);
    snprintf(header_oc.filename, sizeof(header_oc.filename), "%s", file_header);
    if (url_fopen(&header_oc.pb, file_header, URL_WRONLY) < 0) {
            fprintf(stderr, "Could not open '%s'\n", file_header);
	    return NULL;
    }
    fprintf(stderr,"write stream header\n");
    /* write the stream header, if any */
    av_write_header(&header_oc);

    fprintf(stderr,"close file\n");
    /* close the output file */
    url_fclose(&header_oc.pb);

    if (stat (header_oc.filename, &buf) != 0 ) {
	    return NULL;
    }
    header_size=buf.st_size;
    fprintf(stderr,"header size : %d\n",header_size);

    if ((headerfd=open(header_oc.filename,O_RDONLY))<0){
	    fprintf(stderr,"cannot open output file : %s\n",header_oc.filename);
	    return NULL;
    }

    if (NULL== (data=malloc (header_size))){
    	fprintf(stderr,"error : cannont allocate data header buffer\n");
    }
    size_read=read(headerfd,data,header_size);
    
    fprintf(stderr,"ok ,reading %d, return buffer.\n",size_read);
    *ret_size=size_read;
    close(headerfd);
    fprintf(stderr,"destroy stream header\n");
    unlink(header_oc.filename);

    fprintf(stderr,"ok return.\n");
    return data;
}

static struct STRTAB mpeg_codec_list [] ={
#if LIBAVCODEC_BUILD >= 4676
  {FFMPEG_MPEG2,     "FFMpeg Mpeg2"},
#endif
  {FFMPEG_MPEG1,     "FFMpeg Mpeg1"},
  {-1, NULL}
};

static struct STRTAB mpeg_audio_codec_list [] ={
  {FFMPEG_MP2, "FFMpeg MP2"},
#ifdef HAVE_FFMPEG_AC3
  {FFMPEG_AC3, "FFMpeg AC3"},
#endif
#ifdef HAVE_FFMPEG_MP3
  {FFMPEG_MP3, "FFMpeg MP3"},
#endif
  {-1, NULL}
};

static struct STRTAB vob_audio_codec_list [] ={
  {FFMPEG_MP2, "FFMpeg MP2"},
#ifdef HAVE_FFMPEG_AC3
  {FFMPEG_AC3, "FFMpeg AC3"},
#endif
  {-1, NULL}
};

static struct STRTAB dv_codec_list [] ={
#if LIBAVCODEC_BUILD >= 4707
  {FFMPEG_DVVIDEO, "FFMpeg DV"},
#endif
  {-1, NULL}
};

static struct STRTAB dv_audio_codec_list [] ={
  {FFMPEG_PCMS16LE, "FFMpeg PCMS16LE"},
  {-1, NULL}
};

static struct STRTAB mpegts_codec_list[] = {
  {FFMPEG_MPEG4,     "FFMpeg Mpeg4"},
#if LIBAVCODEC_BUILD >= 4676
  {FFMPEG_MPEG2,     "FFMpeg Mpeg2"},
#endif
  {FFMPEG_MPEG1,     "FFMpeg Mpeg1"},
#ifdef HAVE_FFMPEG_X264
#if LIBAVCODEC_BUILD >= 4748
  {FFMPEG_X264,     "FFMpeg H.264"},
#endif
#endif
  {-1, NULL}
};

static struct STRTAB mpegts_audio_codec_list [] ={
  {FFMPEG_MP2, "FFMpeg MP2"},
#ifdef HAVE_FFMPEG_MP3
  {FFMPEG_MP3, "FFMpeg MP3"},
#endif
#ifdef HAVE_FFMPEG_AC3
  {FFMPEG_AC3, "FFMpeg AC3"},
#endif
#ifdef HAVE_FFMPEG_AAC
  {FFMPEG_AAC, "FFMpeg AAC"},
#endif
  {-1, NULL}
};

static struct STRTAB asf_codec_list [] ={
  {FFMPEG_MSMPEG4V2, "FFMpeg MS Mpeg4 V2"},
  {-1, NULL}
};

static struct STRTAB asf_audio_codec_list [] ={
#ifdef HAVE_FFMPEG_MP3
  {FFMPEG_MP3, "FFMpeg MP3"},
#endif
  {-1, NULL}
};

static struct STRTAB mov_codec_list [] ={
  {FFMPEG_MPEG4,     "FFMpeg Mpeg4"},
#if LIBAVCODEC_BUILD >= 4676
  {FFMPEG_MPEG2,     "FFMpeg Mpeg2"},
#endif
  {FFMPEG_MPEG1,     "FFMpeg Mpeg1"},
  {FFMPEG_MJPEG,     "FFMpeg Mjpeg"},
#if LIBAVCODEC_BUILD >= 4670
  {FFMPEG_FLV1,      "FFMpeg Flash Video"},
#endif
#ifdef HAVE_FFMPEG_X264
#if LIBAVCODEC_BUILD >= 4748
  {FFMPEG_X264,     "FFMpeg H.264"},
#endif
#endif
  {FFMPEG_H263P,     "FFMpeg H.263+"},
  {FFMPEG_H263,      "FFMpeg H.263"},
  {FFMPEG_MSMPEG4V2, "FFMpeg MS Mpeg4 V2"},
/* appeared with the 4719 version
but don't work correctly actually */
#if LIBAVCODEC_BUILD >= 16777215
  {FFMPEG_SVQ1,      "FFMpeg SVQ"},
#endif
  {-1, NULL}
};

static struct STRTAB mov_audio_codec_list [] ={
#ifdef HAVE_FFMPEG_MP3
  {FFMPEG_MP3, "FFMpeg MP3"},
#endif
  {FFMPEG_MP2, "FFMpeg MP2"},
#ifdef HAVE_FFMPEG_AC3
  {FFMPEG_AC3, "FFMpeg AC3"},
#endif
#ifdef HAVE_FFMPEG_AAC
  {FFMPEG_AAC, "FFMpeg AAC"},
#endif
  {-1, NULL}
};

static struct STRTAB mp4_codec_list [] ={
  {FFMPEG_MPEG4,     "FFMpeg Mpeg4"},
#if LIBAVCODEC_BUILD >= 4676
  {FFMPEG_MPEG2,     "FFMpeg Mpeg2"},
#endif
  {FFMPEG_MPEG1,     "FFMpeg Mpeg1"},
  {FFMPEG_MJPEG,     "FFMpeg Mjpeg"},
#if LIBAVCODEC_BUILD >= 4670
  {FFMPEG_FLV1,      "FFMpeg Flash Video"},
#endif
#ifdef HAVE_FFMPEG_X264
#if LIBAVCODEC_BUILD >= 4748
  {FFMPEG_X264,     "FFMpeg H.264"},
#endif
#endif
  {FFMPEG_H263P,     "FFMpeg H.263+"},
  {FFMPEG_H263,      "FFMpeg H.263"},
  {FFMPEG_MSMPEG4V2, "FFMpeg MS Mpeg4 V2"},
/* appeared with the 4719 version
but don't work correctly actually */
#if LIBAVCODEC_BUILD >= 16777215
  {FFMPEG_SVQ1,      "FFMpeg SVQ"},
#endif
  {-1, NULL}
};

static struct STRTAB mp4_audio_codec_list [] ={
#ifdef HAVE_FFMPEG_MP3
  {FFMPEG_MP3, "FFMpeg MP3"},
#endif
  {FFMPEG_MP2, "FFMpeg MP2"},
#ifdef HAVE_FFMPEG_AAC
  {FFMPEG_AAC, "FFMpeg AAC"},
#endif
  {-1, NULL}
};

static struct STRTAB _3gp_codec_list [] ={
#ifdef HAVE_FFMPEG
  {FFMPEG_H263,      "FFMpeg H.263"},
  {FFMPEG_MPEG4,     "FFMpeg Mpeg4"},
#endif
  {-1, NULL}
};

static struct STRTAB _3gp_audio_codec_list [] ={
#ifdef HAVE_FFMPEG_AMR_NB
  {FFMPEG_AMR_NB, "FFMpeg AMR NB"},
#endif
#ifdef HAVE_FFMPEG_AMR_WB
  {FFMPEG_AMR_WB, "FFMpeg AMR WB"},
#endif
#ifdef HAVE_FFMPEG_AAC
  {FFMPEG_AAC, "FFMpeg AAC"},
#endif
#ifdef HAVE_FFMPEG_MP3
  {FFMPEG_MP3, "FFMpeg MP3"},
#endif
  {-1, NULL}
};

static struct STRTAB _3g2_codec_list [] ={
#ifdef HAVE_FFMPEG
  {FFMPEG_MPEG4,     "FFMpeg Mpeg4"},
#endif
  {-1, NULL}
};

static struct STRTAB _3g2_audio_codec_list [] ={
#ifdef HAVE_FFMPEG_AAC
  {FFMPEG_AAC, "FFMpeg AAC"},
#endif
#ifdef HAVE_FFMPEG_MP3
  {FFMPEG_MP3, "FFMpeg MP3"},
#endif
  {-1, NULL}
};

static struct STRTAB swf_codec_list [] ={
#if LIBAVCODEC_BUILD >= 4670
  {FFMPEG_FLV1,      "FFMpeg Flash Video"},
#endif
  {FFMPEG_MJPEG,     "FFMpeg Mjpeg"},
  {-1, NULL}
};

static struct STRTAB swf_audio_codec_list [] ={
#ifdef HAVE_FFMPEG_MP3
  {FFMPEG_MP3, "FFMpeg MP3"},
#endif
  {-1, NULL}
};

static struct STRTAB rm_codec_list [] ={
#if LIBAVCODEC_BUILD >= 4735
  {FFMPEG_RV20,      "FFMpeg RealVideo 2.0"},
#endif
  {FFMPEG_RV10,      "FFMpeg RealVideo 1.0"},
  {-1, NULL}
};

static struct STRTAB rm_audio_codec_list [] ={
#ifdef HAVE_FFMPEG_AC3
  {FFMPEG_AC3, "FFMpeg AC3"},
#endif
  {-1, NULL}
};


struct CONTAINER cont_ffmpeg_mpeg = {
	"MPEG",
	"mpeg",
	MPEG_FFMPEG,
	"mpg",
	mpeg_codec_list,
	mpeg_audio_codec_list,
	NULL,
	open_ofile, 
	set_audio,
	set_video,
	write_frame,
	write_audio,
	close_ofile,
	get_header,
	NULL

};
  
struct CONTAINER cont_ffmpeg_mpegts = {
	"MPEGTS",
	"mpegts",
	MPEGTS_FFMPEG,
	"ts",
	mpegts_codec_list,
	mpegts_audio_codec_list,
	NULL,
	open_ofile, 
	set_audio,
	set_video,
	write_frame,
	write_audio,
	close_ofile,
	get_header,
	NULL

};
  
struct CONTAINER cont_ffmpeg_asf = {
	"ASF",
	"asf",
	ASF_FFMPEG,
	"asf",
	asf_codec_list,
	asf_audio_codec_list,
	NULL,
	open_ofile, 
	set_audio,
	set_video,
	write_frame,
	write_audio,
	close_ofile,
	get_header,
	NULL

};
 
struct CONTAINER cont_ffmpeg_vob = {
	"VOB",
	"vob",
	VOB_FFMPEG,
	"vob",
	mpeg_codec_list,
	vob_audio_codec_list,
	NULL,
	open_ofile, 
	set_audio,
	set_video,
	write_frame,
	write_audio,
	close_ofile,
	get_header,
	NULL

};
 
struct CONTAINER cont_ffmpeg_mov = {
	"MOV",
	"mov",
	MOV_FFMPEG,
	"mov",
	mov_codec_list,
	mov_audio_codec_list,
	NULL,
	open_ofile, 
	set_audio,
	set_video,
	write_frame,
	write_audio,
	close_ofile,
	get_header,
	NULL

};
 
struct CONTAINER cont_ffmpeg_dv = {
	"DV",
	"dv",
	DV_FFMPEG,
	"dv",
	dv_codec_list,
	dv_audio_codec_list,
	NULL,
	open_ofile, 
	set_audio,
	set_video,
	write_frame,
	write_audio,
	close_ofile,
	get_header,
	NULL

};
 
struct CONTAINER cont_ffmpeg_mp4 = {
	"MP4",
	"mp4",
	MP4_FFMPEG,
	"mp4",
	mp4_codec_list,
	mp4_audio_codec_list,
	NULL,
	open_ofile, 
	set_audio,
	set_video,
	write_frame,
	write_audio,
	close_ofile,
	get_header,
	NULL

};
 
struct CONTAINER cont_ffmpeg_3gp = {
	"3GP",
	"3gp",
	_3GP_FFMPEG,
	"3gp",
	_3gp_codec_list,
	_3gp_audio_codec_list,
	NULL,
	open_ofile, 
	set_audio,
	set_video,
	write_frame,
	write_audio,
	close_ofile,
	get_header,
	NULL

};
 
struct CONTAINER cont_ffmpeg_3g2 = {
	"3G2",
	"3g2",
	_3G2_FFMPEG,
	"3g2",
	_3g2_codec_list,
	_3g2_audio_codec_list,
	NULL,
	open_ofile, 
	set_audio,
	set_video,
	write_frame,
	write_audio,
	close_ofile,
	get_header,
	NULL

};
 
struct CONTAINER cont_ffmpeg_swf = {
	"SWF",
	"swf",
	SWF_FFMPEG,
	"swf",
	swf_codec_list,
	swf_audio_codec_list,
	NULL,
	open_ofile, 
	set_audio,
	set_video,
	write_frame,
	write_audio,
	close_ofile,
	get_header,
	NULL

};
 
struct CONTAINER cont_ffmpeg_rm = {
	"RM",
	"rm",
	RM_FFMPEG,
	"rm",
	rm_codec_list,
	rm_audio_codec_list,
	NULL,
	open_ofile, 
	set_audio,
	set_video,
	write_frame,
	write_audio,
	close_ofile,
	get_header,
	NULL
};
#endif //LIBAVCODEC_BUILD >= 4754
#endif //HAVE FFMPEG
